/* sequences_publisher.cxx

   A publication of data of type sequences

   This file is derived from code automatically generated by the rtiddsgen 
   command:

   rtiddsgen -language C++ -example <arch> sequences.idl

   Example publication of type sequences automatically generated by 
   'rtiddsgen'. To test them follow these steps:

   (1) Compile this file and the example subscription.

   (2) Start the subscription on the same domain used for RTI Data Distribution
       with the command
       objs/<arch>/sequences_subscriber <domain_id> <sample_count>
                
   (3) Start the publication on the same domain used for RTI Data Distribution
       with the command
       objs/<arch>/sequences_publisher <domain_id> <sample_count>

   (4) [Optional] Specify the list of discovery initial peers and 
       multicast receive addresses via an environment variable or a file 
       (in the current working directory) called NDDS_DISCOVERY_PEERS. 
       
   You can run any number of publishers and subscribers programs, and can 
   add and remove them dynamically from the domain.

                                   
   Example:
        
       To run the example application on domain <domain_id>:
                          
       On Unix: 
       
       objs/<arch>/sequences_publisher <domain_id> o
       objs/<arch>/sequences_subscriber <domain_id> 
                            
       On Windows:
       
       objs\<arch>\sequences_publisher <domain_id>  
       objs\<arch>\sequences_subscriber <domain_id>    

       
modification history
------------ -------       
*/

#include <stdio.h>
#include <stdlib.h>
#include "sequences.h"
#include "sequencesSupport.h"
#include "ndds/ndds_cpp.h"

class WriterListener : public DDSDataWriterListener {


	    /*e \dref_DataWriterListener_on_offered_deadline_missed
     */
    virtual void on_offered_deadline_missed(DDSDataWriter* writer, 
		const DDS_OfferedDeadlineMissedStatus& status){}

    /*e \dref_DataWriterListener_on_liveliness_lost
     */
    virtual void on_liveliness_lost(DDSDataWriter* writer, 
		const DDS_LivelinessLostStatus& status) {}

    /*e \dref_DataWriterListener_on_offered_incompatible_qos
     */
    virtual void on_offered_incompatible_qos(DDSDataWriter* writer, 
		const DDS_OfferedIncompatibleQosStatus& status) {
	    printf("On offered incompatible Qos\n");
	}

    /*e \dref_DataWriterListener_on_publication_matched
     */
    virtual void on_publication_matched(DDSDataWriter* writer, 
		const DDS_PublicationMatchedStatus& status) {
			printf("On Publication matched-> Now start writer\n");
			mGuard->set_trigger_value(DDS_BOOLEAN_TRUE);
	}

    /*e \dref_DataWriterListener_on_reliable_writer_cache_changed
     */
    virtual void on_reliable_writer_cache_changed(DDSDataWriter* writer,
		const DDS_ReliableWriterCacheChangedStatus& status) {}

    /*e \dref_DataWriterListener_on_reliable_reader_activity_changed
     */
    virtual void on_reliable_reader_activity_changed(DDSDataWriter* writer,
		const DDS_ReliableReaderActivityChangedStatus& status) {}

public:
	DDSGuardCondition* mGuard;

	WriterListener ()
	{
		mGuard = new DDSGuardCondition();
	};

	~WriterListener ()
	{
		delete mGuard;
	};



};

/* Delete all entities */
static int publisher_shutdown(
    DDSDomainParticipant *participant)
{
    DDS_ReturnCode_t retcode;
    int status = 0;

    if (participant != NULL) {
        retcode = participant->delete_contained_entities();
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_contained_entities error %d\n", retcode);
            status = -1;
        }

        retcode = DDSTheParticipantFactory->delete_participant(participant);
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_participant error %d\n", retcode);
            status = -1;
        }
    }

    /* RTI Data Distribution Service provides finalize_instance() method on
       domain participant factory for people who want to release memory used
       by the participant factory. Uncomment the following block of code for
       clean destruction of the singleton. */
/*
    retcode = DDSDomainParticipantFactory::finalize_instance();
    if (retcode != DDS_RETCODE_OK) {
        printf("finalize_instance error %d\n", retcode);
        status = -1;
    }
*/

    return status;
}

extern "C" int publisher_main(int domainId, int sample_count, int gather_send_buffer_count_max)
{
    DDSDomainParticipant *participant = NULL;
    DDSPublisher *publisher = NULL;
    DDSTopic *topic = NULL;
    DDSDataWriter *writer = NULL;
    sequencesDataWriter * sequences_writer = NULL;
    sequences *instance = NULL;
    DDS_ReturnCode_t retcode;
    DDS_InstanceHandle_t instance_handle = DDS_HANDLE_NIL;
    const char *type_name = NULL;
    int count = 0;  
    DDS_Duration_t send_period = {4,0};
    const int NumberOfInstances = 32;

    /* To customize participant QoS, use 
       the configuration file USER_QOS_PROFILES.xml */

	DDS_DomainParticipantQos participantQos;
    DDSTheParticipantFactory->get_default_participant_qos(participantQos);
	//

	if ( gather_send_buffer_count_max != 0 )
	{
		//Max gather = 128, default = 16. 16 gives around 5-6 samples aggregated
	  char str[10];
	  sprintf_s(str,9,"%d", gather_send_buffer_count_max);
	  printf("Gather send buffer count = %s\n", str);
	  DDSPropertyQosPolicyHelper::add_property(
		  participantQos.property,
		  "dds.transport.UDPv4.builtin.parent.gather_send_buffer_count_max",
		  str,
		  false
		);
	}
     //

	DDSPropertyQosPolicyHelper::add_property(
		participantQos.property,
		"dds.transport.UDPv4.builtin.parent.message_size_max",
		"65535",
		false
		);

	DDSPropertyQosPolicyHelper::add_property(
		participantQos.property,
		"dds.transport.UDPv4.builtin.send_socket_buffer_size",
		"65535",
		false
		);

	DDSPropertyQosPolicyHelper::add_property(
		participantQos.property,
		"dds.transport.UDPv4.builtin.recv_socket_buffer_size",
		"65535",
		false
		);

	participantQos.receiver_pool.buffer_size = 65535;





    participant = DDSTheParticipantFactory->create_participant(
        domainId, participantQos, 
        NULL /* listener */, DDS_STATUS_MASK_NONE);
    if (participant == NULL) {
        printf("create_participant error\n");
        publisher_shutdown(participant);
        return -1;
    }

    /* To customize publisher QoS, use 
       the configuration file USER_QOS_PROFILES.xml */
    publisher = participant->create_publisher(
        DDS_PUBLISHER_QOS_DEFAULT, NULL /* listener */, DDS_STATUS_MASK_NONE);
    if (publisher == NULL) {
        printf("create_publisher error\n");
        publisher_shutdown(participant);
        return -1;
    }

    /* Register type before creating topic */
    type_name = sequencesTypeSupport::get_type_name();
    retcode = sequencesTypeSupport::register_type(
        participant, type_name);
    if (retcode != DDS_RETCODE_OK) {
        printf("register_type error %d\n", retcode);
        publisher_shutdown(participant);
        return -1;
    }

    /* To customize topic QoS, use 
       the configuration file USER_QOS_PROFILES.xml */
    topic = participant->create_topic(
        "Example sequences",
        type_name, DDS_TOPIC_QOS_DEFAULT, NULL /* listener */,
        DDS_STATUS_MASK_NONE);
    if (topic == NULL) {
        printf("create_topic error\n");
        publisher_shutdown(participant);
        return -1;
    }

    /* To customize data writer QoS, use 
       the configuration file USER_QOS_PROFILES.xml */
    DDS_DataWriterQos dataWriterQos;
    publisher->get_default_datawriter_qos(dataWriterQos);

    dataWriterQos.history.kind = DDS_KEEP_LAST_HISTORY_QOS;

    dataWriterQos.history.depth = 1;

    dataWriterQos.reliability.kind = DDS_RELIABLE_RELIABILITY_QOS;

    //I. We are only using one sample per key. Therefore we only need max_samples-per_instance = 1;
    dataWriterQos.resource_limits.initial_instances =
    dataWriterQos.resource_limits.max_instances = NumberOfInstances;
    dataWriterQos.resource_limits.initial_samples =  
    dataWriterQos.resource_limits.max_samples  = NumberOfInstances;
    //Fix to correct batch size. 
    dataWriterQos.resource_limits.max_samples_per_instance = 1;

    /* Set immediate reader repair response time */
    dataWriterQos.protocol.rtps_reliable_writer.max_nack_response_delay.nanosec = 0;
    dataWriterQos.protocol.rtps_reliable_writer.max_nack_response_delay.sec = 0;

    dataWriterQos.protocol.rtps_reliable_writer.fast_heartbeat_period.sec = 0;
    dataWriterQos.protocol.rtps_reliable_writer.fast_heartbeat_period.nanosec = 50 * 1000000; /* 50 ms */
    dataWriterQos.protocol.rtps_reliable_writer.heartbeat_period.sec = 0;
    dataWriterQos.protocol.rtps_reliable_writer.heartbeat_period.nanosec = 50 * 1000000; /* 50 ms */

    // (later_joiner_hearbeat_period <= hearbeat_period)
    dataWriterQos.protocol.rtps_reliable_writer.late_joiner_heartbeat_period.nanosec = 50 * 1000000; /* 50 ms */
    dataWriterQos.protocol.rtps_reliable_writer.late_joiner_heartbeat_period.sec = 0;

    dataWriterQos.protocol.push_on_write = DDS_BOOLEAN_FALSE;

    //No Piggybacks.
    dataWriterQos.protocol.rtps_reliable_writer.
	    heartbeats_per_max_samples = 0;

    WriterListener* listener = new WriterListener();

    writer = publisher->create_datawriter(
        topic, dataWriterQos, listener /* listener */,
        DDS_STATUS_MASK_ALL);
    if (writer == NULL) {
        printf("create_datawriter error\n");
        publisher_shutdown(participant);
        return -1;
    }
    sequences_writer = sequencesDataWriter::narrow(writer);
    if (sequences_writer == NULL) {
        printf("DataWriter narrow error\n");
        publisher_shutdown(participant);
        return -1;
    }

	//Set wait condition and wait for guard to trigger;
	DDSWaitSet* waitSet = new DDSWaitSet();
	waitSet->attach_condition(listener->mGuard);

    DDSConditionSeq activeConditions;
	retcode = DDS_RETCODE_TIMEOUT;
	while ( retcode == DDS_RETCODE_TIMEOUT)
	{
	  retcode = waitSet->wait(activeConditions, send_period);
	  printf("Waiting on pub match return code error -> %d\n", retcode);

	}
	if ( retcode != DDS_RETCODE_OK)
	{
		printf("Wait set return code error -> %d\n", retcode);
	    publisher_shutdown(participant);
        return -1;
	}
	 printf(" Got publisher match - so start blasting data\n");
   

    /* Create data sample for writing */
    instance = sequencesTypeSupport::create_data();
	sequences* sampleList[100];
	DDS_InstanceHandle_t handles_to_send[NumberOfInstances];
	for ( int i = 0 ;i < NumberOfInstances ; i++)
	{
      sampleList[i] = sequencesTypeSupport::create_data();

	  if (sampleList[i] == NULL) {
        printf("sequencesTypeSupport::create_data error\n");
        publisher_shutdown(participant);
        return -1;
      }

	  /* For a data type that has a key, if the same instance is going to be
       written multiple times, initialize the key here
       and register the keyed instance prior to writing */
      
	  sampleList[i]->id = i + 1;
	  sampleList[i]->count = 0;
	  sampleList[i]->data.length(50);
      handles_to_send[i] = sequences_writer->register_instance(*sampleList[i]);
	} // end loop


    /* Main loop */
    for (count=0; (sample_count == 0) || (count < sample_count); ++count) {

		// Inner loop
        for ( int j = 0; j < NumberOfInstances ; j ++)
		{
          //printf("Writing sequences, count %d\n", count);

          /* Modify the data to be sent here */
        

          retcode = sequences_writer->write(*sampleList[j], handles_to_send[j]);
          if (retcode != DDS_RETCODE_OK) {
            printf("write error %d\n", retcode);
          }

		}
		DDS_ReturnCode_t waitRetCode = sequences_writer->wait_for_acknowledgments(send_period);
		if (waitRetCode == DDS_RETCODE_TIMEOUT)
		{
			printf("writer timed out!\n");
		}//

        //NDDSUtility::sleep(send_period);
    }

/*
    retcode = sequences_writer->unregister_instance(
        *instance, instance_handle);
    if (retcode != DDS_RETCODE_OK) {
        printf("unregister instance error %d\n", retcode);
    }
*/

    /* Delete data sample */
    retcode = sequencesTypeSupport::delete_data(instance);
    if (retcode != DDS_RETCODE_OK) {
        printf("sequencesTypeSupport::delete_data error %d\n", retcode);
    }

    /* Delete all entities */
    return publisher_shutdown(participant);
}

#if defined(RTI_WINCE)
int wmain(int argc, wchar_t** argv)
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */ 
    
    if (argc >= 2) {
        domainId = _wtoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = _wtoi(argv[2]);
    }

     /* Uncomment this to turn on additional logging
    NDDSConfigLogger::get_instance()->
        set_verbosity_by_category(NDDS_CONFIG_LOG_CATEGORY_API, 
                                  NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */
    
    return publisher_main(domainId, sample_count);
}
 
#elif !(defined(RTI_VXWORKS) && !defined(__RTP__)) && !defined(RTI_PSOS)
int main(int argc, char *argv[])
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */
	int gather_send_buffer_count_max = 0;

    if (argc >= 2) {
        domainId = atoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = atoi(argv[2]);
    }

    if (argc >= 4) {
        gather_send_buffer_count_max = atoi(argv[3]);
    }

    /* Uncomment this to turn on additional logging 
    NDDSConfigLogger::get_instance()->
        set_verbosity(NDDS_CONFIG_LOG_VERBOSITY_WARNING); */
    
    
    return publisher_main(domainId, sample_count, gather_send_buffer_count_max);
}
#endif
