/* sequences_subscriber.cxx

   A subscription example

   This file is derived from code automatically generated by the rtiddsgen 
   command:

   rtiddsgen -language C++ -example <arch> sequences.idl

   Example subscription of type sequences automatically generated by 
   'rtiddsgen'. To test them follow these steps:

   (1) Compile this file and the example publication.

   (2) Start the subscription on the same domain used for RTI Data Distribution
       Service with the command
       objs/<arch>/sequences_subscriber <domain_id> <sample_count>

   (3) Start the publication on the same domain used for RTI Data Distribution
       with the command
       objs/<arch>/sequences_publisher <domain_id> <sample_count>

   (4) [Optional] Specify the list of discovery initial peers and 
       multicast receive addresses via an environment variable or a file 
       (in the current working directory) called NDDS_DISCOVERY_PEERS. 
       
   You can run any number of publishers and subscribers programs, and can 
   add and remove them dynamically from the domain.
              
                                   
   Example:
        
       To run the example application on domain <domain_id>:
                          
       On Unix: 
       
       objs/<arch>/sequences_publisher <domain_id> 
       objs/<arch>/sequences_subscriber <domain_id> 
                            
       On Windows:
       
       objs\<arch>\sequences_publisher <domain_id>  
       objs\<arch>\sequences_subscriber <domain_id>   
              
       
modification history
------------ -------       
*/

#include <stdio.h>
#include <stdlib.h>
#include "sequences.h"
#include "sequencesSupport.h"
#include "ndds/ndds_cpp.h"

class sequencesListener : public DDSDataReaderListener {
  public:
    virtual void on_requested_deadline_missed(
        DDSDataReader* /*reader*/,
        const DDS_RequestedDeadlineMissedStatus& /*status*/) {}
    
    virtual void on_requested_incompatible_qos(
        DDSDataReader* /*reader*/,
        const DDS_RequestedIncompatibleQosStatus& status) {
	printf("On requested incompatible Qos -> %d\n", status.last_policy_id);}
    
    virtual void on_sample_rejected(
        DDSDataReader* /*reader*/,
        const DDS_SampleRejectedStatus& /*status*/) {
	printf("On Sample rejected\n");}

    virtual void on_liveliness_changed(
        DDSDataReader* /*reader*/,
        const DDS_LivelinessChangedStatus& /*status*/) {
	     printf("on liveliness changed \n");}

    virtual void on_sample_lost(
        DDSDataReader* /*reader*/,
        const DDS_SampleLostStatus& /*status*/) {
	   printf("On sample lost\n");}

    virtual void on_subscription_matched(
        DDSDataReader* /*reader*/,
        const DDS_SubscriptionMatchedStatus& /*status*/) {
			printf("On Subscription Matched\n");}

    virtual void on_data_available(DDSDataReader* reader);

	int mSampleCount;

	sequencesListener()
	{
		mSampleCount = 0;
	};
};

void sequencesListener::on_data_available(DDSDataReader* reader)
{
    sequencesDataReader *sequences_reader = NULL;
    sequencesSeq data_seq;
    DDS_SampleInfoSeq info_seq;
    DDS_ReturnCode_t retcode;
    //int i;

    sequences_reader = sequencesDataReader::narrow(reader);
    if (sequences_reader == NULL) {
        printf("DataReader narrow error\n");
        return;
    }

    retcode = sequences_reader->take(
        data_seq, info_seq, DDS_LENGTH_UNLIMITED,
        DDS_ANY_SAMPLE_STATE, DDS_ANY_VIEW_STATE, DDS_ANY_INSTANCE_STATE);

    if (retcode == DDS_RETCODE_NO_DATA) {
        return;
    } else if (retcode != DDS_RETCODE_OK) {
        printf("take error %d\n", retcode);
        return;
    }

	mSampleCount += data_seq.length();
    //for (i = 0; i < data_seq.length(); ++i) {
        //if (info_seq[i].valid_data) {
            //sequencesTypeSupport::print_data(&data_seq[i]);
       // }
    //}
	//printf("Number of samples received = %d\n",data_seq.length());

    retcode = sequences_reader->return_loan(data_seq, info_seq);
    if (retcode != DDS_RETCODE_OK) {
        printf("return loan error %d\n", retcode);
    }
}

/* Delete all entities */
static int subscriber_shutdown(
    DDSDomainParticipant *participant)
{
    DDS_ReturnCode_t retcode;
    int status = 0;

    if (participant != NULL) {
        retcode = participant->delete_contained_entities();
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_contained_entities error %d\n", retcode);
            status = -1;
        }

        retcode = DDSTheParticipantFactory->delete_participant(participant);
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_participant error %d\n", retcode);
            status = -1;
        }
    }

    /* RTI Data Distribution Service provides the finalize_instance() method on
       domain participant factory and the finalize() method on type support for
       users who want to release memory used by the participant factory and
       type support singletons. Uncomment the following block of code for
       clean destruction of the singletons. */
/*
    sequencesTypeSupport::finalize();

    retcode = DDSDomainParticipantFactory::finalize_instance();
    if (retcode != DDS_RETCODE_OK) {
        printf("finalize_instance error %d\n", retcode);
        status = -1;
    }
*/
    return status;
}

extern "C" int subscriber_main(int domainId, int sample_count)
{
    DDSDomainParticipant *participant = NULL;
    DDSSubscriber *subscriber = NULL;
    DDSTopic *topic = NULL;
    sequencesListener *reader_listener = NULL; 
    DDSDataReader *reader = NULL;
    DDS_ReturnCode_t retcode;
    const char *type_name = NULL;
    int count = 0;
    DDS_Duration_t receive_period = {0,100000000};
    int status = 0;

	const int NumberOfInstances = 100;


    /* To customize participant QoS, use */
	DDS_DomainParticipantQos participantQos;
    DDSTheParticipantFactory->get_default_participant_qos(participantQos);

	DDSPropertyQosPolicyHelper::add_property(
		participantQos.property,
		"dds.transport.UDPv4.builtin.parent.message_size_max",
		"65535",
		false
		);

	DDSPropertyQosPolicyHelper::add_property(
		participantQos.property,
		"dds.transport.UDPv4.builtin.send_socket_buffer_size",
		"65535",
		false
		);

	DDSPropertyQosPolicyHelper::add_property(
		participantQos.property,
		"dds.transport.UDPv4.builtin.recv_socket_buffer_size",
		"65535",
		false
		);

	participantQos.receiver_pool.buffer_size = 65535;

    participant = DDSTheParticipantFactory->create_participant(
        domainId, participantQos, 
        NULL /* listener */, DDS_STATUS_MASK_NONE);
    if (participant == NULL) {
        printf("create_participant error\n");
        subscriber_shutdown(participant);
        return -1;
    }

    /* To customize the subscriber QoS, use 
       the configuration file USER_QOS_PROFILES.xml */
    subscriber = participant->create_subscriber(
        DDS_SUBSCRIBER_QOS_DEFAULT, NULL /* listener */, DDS_STATUS_MASK_NONE);
    if (subscriber == NULL) {
        printf("create_subscriber error\n");
        subscriber_shutdown(participant);
        return -1;
    }

    /* Register the type before creating the topic */
    type_name = sequencesTypeSupport::get_type_name();
    retcode = sequencesTypeSupport::register_type(
        participant, type_name);
    if (retcode != DDS_RETCODE_OK) {
        printf("register_type error %d\n", retcode);
        subscriber_shutdown(participant);
        return -1;
    }

    /* To customize the topic QoS, use 
       the configuration file USER_QOS_PROFILES.xml */
    topic = participant->create_topic(
        "Example sequences",
        type_name, DDS_TOPIC_QOS_DEFAULT, NULL /* listener */,
        DDS_STATUS_MASK_NONE);
    if (topic == NULL) {
        printf("create_topic error\n");
        subscriber_shutdown(participant);
        return -1;
    }

    /* Create a data reader listener */
    reader_listener = new sequencesListener();

	DDS_DataReaderQos dataReaderQos;
	subscriber->get_default_datareader_qos(dataReaderQos);

	dataReaderQos.history.kind = DDS_KEEP_LAST_HISTORY_QOS;
	dataReaderQos.history.depth = 100;

	dataReaderQos.reliability.kind = DDS_RELIABLE_RELIABILITY_QOS;

	dataReaderQos.protocol.rtps_reliable_reader.max_heartbeat_response_delay.sec = 0;
	dataReaderQos.protocol.rtps_reliable_reader.max_heartbeat_response_delay.nanosec = 0;
	dataReaderQos.protocol.rtps_reliable_reader.heartbeat_suppression_duration.nanosec = 0;
	dataReaderQos.protocol.rtps_reliable_reader.heartbeat_suppression_duration.sec = 0;


	dataReaderQos.resource_limits.initial_instances =
		dataReaderQos.resource_limits.max_instances = 100*NumberOfInstances;
	dataReaderQos.resource_limits.initial_samples =
		dataReaderQos.resource_limits.max_samples = 100*NumberOfInstances;
	//Note we are only sending one sample per instance!!!!!!!!!!
	dataReaderQos.resource_limits.max_samples_per_instance = 100;
	dataReaderQos.reader_resource_limits.max_samples_per_remote_writer = NumberOfInstances;


    reader = subscriber->create_datareader(
        topic, dataReaderQos, reader_listener,
        ~DDS_DATA_AVAILABLE_STATUS);
    if (reader == NULL) {
        printf("create_datareader error\n");
        subscriber_shutdown(participant);
        delete reader_listener;
        return -1;
    }

	int lastSampleCount = 0;
    /* Main loop */
    for (count=0; (sample_count == 0) || (count < sample_count); ++count) {

        //printf("sequences subscriber sleeping for %d sec...\n",
        //       receive_period.sec);

        NDDSUtility::sleep(receive_period);
		// Poll Reader
		reader_listener->on_data_available(reader);

		if ( count % 10 == 0 )
		{
			printf("Samples per second = %d\n", reader_listener->mSampleCount - lastSampleCount);
			lastSampleCount = reader_listener->mSampleCount;
		}
    }

    /* Delete all entities */
    status = subscriber_shutdown(participant);
    delete reader_listener;

    return status;
}

#if defined(RTI_WINCE)
int wmain(int argc, wchar_t** argv)
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */ 
    
    if (argc >= 2) {
        domainId = _wtoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = _wtoi(argv[2]);
    }
    
    /* Uncomment this to turn on additional logging
    NDDSConfigLogger::get_instance()->
        set_verbosity_by_category(NDDS_CONFIG_LOG_CATEGORY_API, 
                                  NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */
                                  
    return subscriber_main(domainId, sample_count);
}

#elif !(defined(RTI_VXWORKS) && !defined(__RTP__)) && !defined(RTI_PSOS)
int main(int argc, char *argv[])
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */

    if (argc >= 2) {
        domainId = atoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = atoi(argv[2]);
    }


    /* Uncomment this to turn on additional logging
    NDDSConfigLogger::get_instance()->
        set_verbosity_by_category(NDDS_CONFIG_LOG_CATEGORY_API, 
                                  NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */
                                  
    return subscriber_main(domainId, sample_count);
}
#endif

